from __future__ import annotations

from datetime import datetime
from io import BytesIO
from typing import Any, cast
from unittest import mock as _mock

import factory
import pytest
import pytz
from freezegun import freeze_time
from pytest_factoryboy import register
from responses import RequestsMock
from werkzeug.datastructures import FileStorage

from ckan.lib.redis import connect_to_redis
from ckan.tests import factories
from ckan.tests.helpers import call_action  # pyright: ignore[reportUnknownVariableType]

from ckanext.files import shared

call_action: Any


@pytest.fixture
def mock():
    return _mock


@pytest.fixture
def responses(ckan_config: dict[str, Any]):
    with RequestsMock() as rsps:
        rsps.add_passthru(ckan_config["solr_url"])
        yield rsps


@pytest.fixture
def files_stopped_time():
    now = datetime.now(pytz.utc)
    with freeze_time(now):
        yield now


@pytest.fixture
def clean_db(reset_db: Any, migrate_db_for: Any):
    reset_db()
    migrate_db_for("files")


class FakeFileStorage(FileStorage):
    def __init__(self, stream: Any, filename: str):
        self.filename = filename

        super().__init__(stream, filename, "upload")


@pytest.fixture
def create_with_upload(ckan_config: dict[str, Any], monkeypatch: Any, tmpdir: Any):
    """Reimplementation of original CKAN fixture with better fake storage.

    CKAN version adds just a few attributes to FileStorage object, while
    current plugin requires exact immitation.
    """
    from ckan.lib import uploader

    storage_path = str(tmpdir)
    monkeypatch.setitem(ckan_config, "ckan.storage_path", storage_path)
    if hasattr(uploader, "_storage_path"):
        monkeypatch.setattr(uploader, "_storage_path", storage_path)

    def factory(data: Any, filename: str, context: Any = None, **kwargs: Any):
        if context is None:
            context = {}

        action = kwargs.pop("action", "resource_create")
        field = kwargs.pop("upload_field_name", "upload")
        test_file = BytesIO()
        if isinstance(data, str):
            data = data.encode()
        test_file.write(data)
        test_file.seek(0)
        test_resource = FakeFileStorage(test_file, filename)

        params = {
            field: test_resource,
        }
        params.update(kwargs)
        return call_action(action, context, **params)

    return factory


@pytest.fixture(scope="session")
def reset_redis():
    def cleaner(pattern: str = "*"):
        """Remove keys matching pattern.

        Return number of removed records.
        """
        conn = connect_to_redis()
        keys = cast(Any, conn.keys(pattern))
        if keys:
            return cast(int, conn.delete(*keys))
        return 0

    return cleaner


@pytest.fixture
def clean_redis(reset_redis: Any):
    """Remove all keys from Redis.

    This fixture removes all the records from Redis.

    Example:
        ```python
        @pytest.mark.usefixtures("clean_redis")
        def test_redis_is_empty():
            assert redis.keys("*") == []
        ```

    If test requires presence of some initial data in redis, make sure that
    data producer applied **after** ``clean_redis``:

    Example:
        ```python
        @pytest.mark.usefixtures(
            "clean_redis",
            "fixture_that_adds_xxx_key_to_redis"
        )
        def test_redis_has_one_record():
            assert redis.keys("*") == [b"xxx"]
        ```
    """
    reset_redis()


@register(_name="file")
class FileFactory(factories.CKANFactory):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        model = shared.File
        action = "files_file_create"

    name = factory.LazyFunction(factories.fake.unique.file_name)  # pyright: ignore[reportPrivateImportUsage]
    upload = factory.Faker("binary", length=100)  # pyright: ignore[reportPrivateImportUsage]



"""convert size to bigint.

Revision ID: 565ef5eca492
Revises: d19094d45a0f
Create Date: 2024-12-02 02:40:53.449550

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "565ef5eca492"
down_revision = "d19094d45a0f"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        table_name="files_file",
        column_name="size",
        nullable=False,
        server_default="0",
        type_=sa.BigInteger,
    )
    op.alter_column(
        table_name="files_multipart",
        column_name="size",
        nullable=False,
        server_default="0",
        type_=sa.BigInteger,
    )


def downgrade():
    op.alter_column(
        table_name="files_file",
        column_name="size",
        nullable=False,
        server_default="0",
        type_=sa.Integer,
    )
    op.alter_column(
        table_name="files_multipart",
        column_name="size",
        nullable=False,
        server_default="0",
        type_=sa.Integer,
    )



from __future__ import annotations

import dataclasses
import logging
import os
from typing import Any

import file_keeper as fk
import flask
from file_keeper.default.adapters import fs
from typing_extensions import override

from ckan import types
from ckan.config.declaration import Declaration, Key

from ckanext.files import shared

log = logging.getLogger(__name__)
CHUNK_SIZE = 16384


@dataclasses.dataclass()
class Settings(shared.Settings, fs.Settings):
    pass


class Reader(shared.Reader, fs.Reader):
    @override
    def response(self, data: shared.FileData, extras: dict[str, Any]) -> types.Response:
        filepath = os.path.join(self.storage.settings.path, data.location)
        return flask.send_file(
            filepath,
            download_name=data.location,
            mimetype=data.content_type,
        )


class FsStorage(shared.Storage, fs.FsStorage):  # pyright: ignore[reportIncompatibleVariableOverride]
    """Store files in local filesystem."""

    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory: type[shared.Settings] = Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    ReaderFactory: type[shared.Reader] = Reader  # pyright: ignore[reportIncompatibleVariableOverride]
    UploaderFactory = type("Uploader", (shared.Uploader, fs.Uploader), {})
    ManagerFactory = type("Manager", (shared.Manager, fs.Manager), {})


class PublicFsReader(Reader):
    capabilities = fs.Reader.capabilities | fk.Capability.LINK_PERMANENT
    storage: PublicFsStorage  # pyright: ignore[reportIncompatibleVariableOverride]

    @override
    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        """Return public download link."""
        from ckan.lib.helpers import url_for_static_or_external  # noqa: PLC0415

        return url_for_static_or_external(
            os.path.join(
                self.storage.settings.public_prefix,
                data.location,
            ),
            _external=True,
        )


@dataclasses.dataclass()
class PublicFsSettings(Settings):
    public_prefix: str = "/"


class PublicFsStorage(FsStorage):
    settings: PublicFsSettings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = PublicFsSettings
    ReaderFactory = PublicFsReader

    @override
    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)

        declaration.declare(key.public_prefix, "/").set_description(
            "URL of the storage folder." + " `public_prefix + location` must produce a public URL",
        )


class CkanResourceFsStorage(FsStorage):
    pass



"""empty message.

Revision ID: b4b6ff46cd13
Revises: ea61ab6a6042
Create Date: 2026-04-03 16:38:46.405174

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b4b6ff46cd13"
down_revision = "ea61ab6a6042"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_column("files_transfer_history", "item_type")
    op.alter_column("files_transfer_history", "item_id", new_column_name="file_id")

    op.drop_column("files_owner", "item_type")
    op.alter_column("files_owner", "item_id", new_column_name="file_id")
    op.create_primary_key("files_owner_pkey", "files_owner", ["file_id"])

    op.create_foreign_key(
        "files_owner_file_id_fkey", "files_owner", "files_file", ["file_id"], ["id"], ondelete="CASCADE"
    )

    op.create_foreign_key(
        "files_transfer_history_owner_file_id_fkey",
        "files_transfer_history",
        "files_owner",
        ["file_id"],
        ["file_id"],
        ondelete="CASCADE",
    )


def downgrade():
    op.drop_constraint("files_transfer_history_owner_file_id_fkey", "files_transfer_history", type_="foreignkey")
    op.alter_column("files_transfer_history", "file_id", new_column_name="item_id")
    op.add_column("files_transfer_history", sa.Column("item_type", sa.TEXT(), autoincrement=False, nullable=False))
    op.create_index("idx_item", "files_transfer_history", ["item_id", "item_type"], unique=False)

    op.drop_constraint("files_owner_file_id_fkey", "files_owner", type_="foreignkey")

    op.add_column("files_owner", sa.Column("item_type", sa.TEXT(), autoincrement=False, nullable=False))
    op.alter_column("files_owner", "file_id", new_column_name="item_id")
    op.drop_constraint("files_owner_pkey", "files_owner")
    op.create_primary_key("files_owner_pkey", "files_owner", ["item_id", "item_type"])

    op.create_foreign_key(
        "files_transfer_history_owner_file_id_fkey",
        "files_transfer_history",
        "files_owner",
        ["item_id", "item_type"],
        ["item_id", "item_type"],
    )



from __future__ import annotations

from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, backref, relationship

from ckan.lib.dictization import table_dictize
from ckan.model.types import make_uuid
from ckan.types import Context

from .base import Base, now
from .owner import FilesOwner


class TransferHistory(Base):
    """Model for tracking ownership history of the file.

    Keyword Args:
        file_id (str): ID of the owned object
        owner_id (str): ID of the owner
        owner_type (str): Type of the owner
        leave_date (datetime): date of ownership transfer to a different owner
        actor (str | None): user who initiated ownership transfer

    Example:
        ```python
        record = TransferHistory(
            file_id=file.id,
            owner_id=prev_owner.owner_id,
            owner_type=prev_owner.owner_type,
        )
        ```
    """

    __table__ = sa.Table(
        "files_transfer_history",
        Base.metadata,
        sa.Column("id", sa.Text, primary_key=True, default=make_uuid),
        sa.Column("file_id", sa.Text, nullable=False),
        sa.Column("owner_id", sa.Text, nullable=False),
        sa.Column("owner_type", sa.Text, nullable=False),
        sa.Column(
            "at",
            sa.DateTime(timezone=True),
            default=now,
            nullable=False,
        ),
        sa.Column("action", sa.Text, nullable=False, default="transfer"),
        sa.Column("actor", sa.Text, nullable=False),
        sa.ForeignKeyConstraint(
            ["file_id"],
            ["files_owner.file_id"],
            "files_transfer_history_owner_file_id_fkey",
            ondelete="CASCADE",
        ),
    )
    id: Mapped[str]
    file_id: Mapped[str]
    owner_id: Mapped[str]
    owner_type: Mapped[str]
    at: Mapped[datetime]
    action: Mapped[str]
    actor: Mapped[str]

    current: Mapped[FilesOwner] = relationship(  # type: ignore
        FilesOwner,
        backref=backref("history", cascade="delete, delete-orphan"),
    )

    def dictize(self, context: Context):
        return table_dictize(self, context)

    @classmethod
    def from_owner(cls, owner: FilesOwner, actor: str = ""):
        return cls(
            file_id=owner.file_id,
            owner_id=owner.owner_id,
            owner_type=owner.owner_type,
            actor=actor,
        )



import contextlib

from .db import DbStorage
from .filebin import FilebinStorage
from .fs import CkanResourceFsStorage, FsStorage, PublicFsStorage
from .link import LinkStorage
from .redis import RedisStorage

with contextlib.suppress(ImportError):
    from .s3 import S3Storage

with contextlib.suppress(ImportError):
    from .google_cloud import GoogleCloudStorage

with contextlib.suppress(ImportError):
    from .opendal import OpenDalStorage

with contextlib.suppress(ImportError):
    from .azure import AzureBlobStorage

with contextlib.suppress(ImportError):
    from .libcloud import LibCloudStorage


__all__ = [
    "FsStorage",
    "PublicFsStorage",
    "CkanResourceFsStorage",
    "RedisStorage",
    "S3Storage",
    "GoogleCloudStorage",
    "FilebinStorage",
    "OpenDalStorage",
    "LibCloudStorage",
    "DbStorage",
    "LinkStorage",
    "AzureBlobStorage",
]



from __future__ import annotations

import os
from typing import Any

import click
import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.files import config, shared
from ckanext.files.base import get_storage
from ckanext.files.model import File, Owner


@click.group()
def group():
    """Storage-level operations."""


@group.command("list")
@click.option("-v", "--verbose", is_flag=True, help="Show storage's details")
def list_storages(verbose: bool):
    """Show all configured storages."""
    for name, settings in config.storages().items():
        click.secho("{}: {}".format(click.style(name, bold=True), settings["type"]))
        if verbose:
            storage = shared.get_storage(name)
            click.echo(f"\tSupports: {storage.capabilities}")
            click.echo(f"\tDoes not support: {~storage.capabilities}")


@group.command()
@click.option("-s", "--storage-name", help="Name of the configured storage")
@click.option(
    "-u",
    "--untracked-only",
    help="Show only untracked files(not recorded in DB)",
    is_flag=True,
)
@click.option(
    "-t",
    "--track",
    help="Track untracked files by creating record in DB",
    is_flag=True,
)
@click.option("-a", "--adopt-by", help="Attach untracked to specified user")
def scan(
    storage_name: str | None,
    untracked_only: bool,
    adopt_by: str | None,
    track: bool,
):
    """Iterate over all files available in storage.

    This command can be used to locate untracked files, that are not registered
    in DB, but exist in storage.

    """
    storage_name = storage_name or config.default_storage()
    storage = shared.get_storage(storage_name)

    try:
        files = storage.scan()
    except shared.exc.UnsupportedOperationError as err:
        tk.error_shout(err)
        raise click.Abort from err

    stepfather = model.User.get(adopt_by)

    for name in files:
        is_untracked = not model.Session.query(
            File.by_location(name, storage_name).exists(),
        ).scalar()

        if untracked_only and not is_untracked:
            continue

        click.echo(name)

        if track and is_untracked:
            try:
                data = storage.analyze(shared.Location(name))
            except shared.exc.UnsupportedOperationError as err:
                tk.error_shout(err)
                raise click.Abort from err

            fileobj = File(
                name=os.path.basename(name),
                storage=storage_name,
            )
            data.into_object(fileobj)

            model.Session.add(fileobj)

            if stepfather:
                owner = Owner(
                    file_id=fileobj.id,
                    owner_id=stepfather.id,
                    owner_type="user",
                )
                model.Session.add(owner)

            model.Session.commit()


@group.command()
@click.argument("src")
@click.argument("dest")
@click.option(
    "-i",
    "--id",
    help="IDs of files for transfer",
    multiple=True,
)
@click.option(
    "-r",
    "--remove",
    help="Remove file from the source after transfer",
    is_flag=True,
)
def transfer(src: str, dest: str, id: tuple[str, ...], remove: bool):
    """Move files between storages."""
    from_storage = shared.get_storage(src)
    to_storage = shared.get_storage(dest)

    is_supported = from_storage.supports_synthetic(
        shared.Capability.MOVE if remove else shared.Capability.COPY, to_storage
    )

    if not is_supported:
        tk.error_shout("Operation is not supported")
        raise click.Abort

    op = from_storage.move_synthetic if remove else from_storage.copy_synthetic

    stmt = sa.select(shared.File).where(shared.File.storage == src)
    if id:
        stmt = stmt.where(shared.File.id.in_(id))

    total = model.Session.scalar(stmt.with_only_columns(sa.func.count()))
    files = model.Session.scalars(stmt)

    with click.progressbar(files, length=total) as bar:
        for item in bar:
            data = shared.FileData.from_object(item)
            new_data = op(data.location, data, to_storage)
            new_data.into_object(item)
            item.storage = dest
            model.Session.commit()


@group.command()
@click.option("-s", "--storage-name", help="Name of the configured storage")
@click.option("-y", "--yes", help="Do not ask for confirmation", is_flag=True)
def clean(storage_name: str | None, yes: bool):
    """Remove all tracked files from the storage.

    Untracked files are not removed, so you may want to run `storage scan -t`
    first.
    """
    user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    storage_name = storage_name or config.default_storage()
    if not get_storage(storage_name).supports(shared.Capability.REMOVE):
        tk.error_shout(f"Storage {storage_name} does not support file removal")
        raise click.Abort

    search = tk.get_action("files_file_search")
    overview = search({"user": user["name"]}, {"storage": storage_name, "rows": 0})

    click.echo(f"Storage {storage_name} contains {overview['count']} files.")

    if not overview["count"]:
        click.secho("Done", fg="green")

    if not yes and not click.confirm("Proceed with removal?"):
        return

    bar: Any = click.progressbar(length=overview["count"])
    with bar:
        while True:
            result = search({"user": user["name"]}, {"storage": storage_name})
            if not result["count"]:
                break

            for item in result["results"]:
                bar.label = f"Removing {item['id']}"
                tk.get_action("files_file_delete")({"user": user["name"]}, {"id": item["id"]})
                bar.update(1)

    click.secho("Done", fg="green")



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import dataclasses
from time import time
from typing import Any, ClassVar, cast

import file_keeper as fk
import flask
from typing_extensions import TypeAlias, override

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.config.declaration import Declaration, Key

from . import config, utils

adapters = fk.adapters
storages = fk.Registry["fk.Storage"]()

FileData: TypeAlias = fk.FileData

make_storage = fk.make_storage


def get_storage(name: str | None = None) -> fk.Storage:
    """Return existing storage instance.

    Storages are initialized when plugin is loaded. As result, this function
    always returns the same storage object for the given name.

    If no name specified, default storage is returned.

    Args:
        name: name of the configured storage

    Returns:
        storage instance

    Raises:
        UnknownStorageError: storage with the given name is not configured

    Example:
        ```
        default_storage = get_storage()
        storage = get_storage("storage name")
        ```

    """
    if name is None:
        name = config.default_storage()

    storage = storages.get(name)

    if not storage:
        raise fk.exc.UnknownStorageError(name)

    return storage


@dataclasses.dataclass()
class Settings(fk.Settings):
    supported_types: list[str] = cast("list[str]", dataclasses.field(default_factory=list))
    max_size: int = -1
    public: bool = False
    """Whether storage is public and allows unauthenticated access."""


class Uploader(fk.Uploader):
    pass


class Manager(fk.Manager):
    pass


class Reader(fk.Reader):
    """Service responsible for reading data from the storage.

    :py:class:`Storage` internally calls methods of this service. For example,
    ``Storage.stream(data, **kwargs)`` results in ``Reader.stream(data,
    kwargs)``.

    >>> class MyReader(Reader):
    >>>     def stream(
    >>>         self, data: FileData, extras: dict[str, Any]
    >>>     ) -> Iterable[bytes]:
    >>>         return open(data.location, "rb")

    """

    def response(self, data: FileData, extras: dict[str, Any]) -> types.Response:
        if not self.storage.supports(fk.Capability.STREAM):
            raise fk.exc.UnsupportedOperationError("stream", self)

        return flask.Response(
            self.stream(data, extras),
            mimetype=data.content_type or None,
            headers={"Content-length": str(data.size)},
        )


class Storage(fk.Storage):
    """Base class for storage implementation."""

    settings: Settings
    reader: Reader
    uploader: Uploader
    manager: Manager

    SettingsFactory: ClassVar[type[Settings]] = Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    UploaderFactory: ClassVar[type[Uploader]] = Uploader  # pyright: ignore[reportIncompatibleVariableOverride]
    ReaderFactory: ClassVar[type[Reader]] = Reader  # pyright: ignore[reportIncompatibleVariableOverride]
    ManagerFactory: ClassVar[type[Manager]] = Manager  # pyright: ignore[reportIncompatibleVariableOverride]

    def validate_size(self, size: int):
        max_size = self.settings.max_size
        if max_size >= 0 and size > max_size:
            raise fk.exc.LargeUploadError(size, max_size)

    def validate_content_type(self, content_type: str):
        supported_types = self.settings.supported_types
        if supported_types and not utils.is_supported_type(
            content_type,
            supported_types,
        ):
            raise fk.exc.WrongUploadTypeError(content_type)

    @override
    def upload(self, location: fk.Location, upload: fk.Upload, /, **kwargs: Any) -> FileData:
        self.validate_size(upload.size)
        self.validate_content_type(upload.content_type)

        return super().upload(location, upload, **kwargs)

    @override
    def multipart_start(self, location: fk.Location, size: int, /, **kwargs: Any) -> FileData:
        self.validate_size(size)
        self.validate_content_type(kwargs.get("content_type", ""))

        return super().multipart_start(location, size, **kwargs)

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        declaration.declare(key.max_size, -1).append_validators(
            "files_parse_filesize",
        ).set_description(
            "The maximum size of a single upload."
            + "\nSupports size suffixes: 42B, 2M, 24KiB, 1GB."
            + " `0` means no restrictions.",
        )

        declaration.declare_list(key.supported_types, None).set_description(
            "Space-separated list of MIME types or just type or subtype part."
            + "\nExample: text/csv pdf application video jpeg",
        )

        declaration.declare_bool(key.public).set_description(
            "Whether storage is public and allows unauthenticated access."
        )

        declaration.declare_bool(key.overwrite_existing, True).set_description(
            "If file already exists, replace it with new content.",
        )

        declaration.declare_bool(key.initialize).set_description(
            "Prepare storage backend for uploads(create path, bucket, DB)."
            + " This option depends on\nthe adapter and can be ignored if an adapter"
            + " cannot safely initialize the storage path.\nAlways prefer manual"
            + " creation of location specified in the ``path`` option.",
        )

        declaration.declare(key.path, "").set_description(
            "Prefix for the file's location. The actual meaning of this option"
            + " depends on the adapter. FS adapter uses the path as a root folder for uploads."
            + " Cloud adapters, usually, use path as a prefix for the name of uploaded objects.",
        )

        declaration.declare(key.name, key[-1]).set_description(
            "Descriptive name of the storage used for debugging. When empty,"
            + " name from the config option is used,"
            + " i.e: `ckanext.files.storage.DEFAULT_NAME...`",
        )

        declaration.declare(key.hashing_algorithm, "md5").set_description(
            "Hashing algorithm used to calculate file's hash. This option is used by"
            " uploaders to calculate\nfile's hash and store it in FileData. Supported"
            " values depend on the implementation\nof the uploader, but common"
            " algorithms are: `md5`, `sha1`, `sha256`.\nIf storage adapter does not"
            " support customization of algorithm, it will ignore this option.\nFileData"
            " object produced by storage's ``upload`` method should contain the actual"
            " algorithm\nused to compute the hash."
        )

        declaration.declare_list(key.location_transformers, None).set_description(
            "List of transformations applied to the file location."
            " Depending on the storage type, sanitizing the path or removing"
            " special characters can be sensible. Empty value leaves location"
            " unchanged, `uuid` transforms location into UUID, `uuid_with_extension`"
            " transforms filename into UUID and appends original file's extension"
            " to it.",
        )

        declaration.declare_list(key.disabled_capabilities, None).set_description(
            "Capabilities that are not supported even if implemented."
            " Can be used to transform fully-featured storage into a"
            " read-only/write-only storage."
        )

    def as_response(
        self,
        data: FileData,
        filename: str | None = None,
        /,
        send_inline: bool = False,
        **kwargs: Any,
    ) -> types.Response:
        """Make Flask response with file attachment.

        By default, files are served as attachments and are downloaded as
        result.

        If rendering is safe and preferable enable ``send_inline`` flag.

        Args:
            data: file details
            filename: expected name of the file used instead of the real name

        Keyword Args:
            send_inline: do not force download and try rendering file in browser
            **kwargs: ...

        Returns:
            Flask response with file's content
        """
        try:
            resp = self.reader.response(data, kwargs)
        except fk.exc.MissingFileError:
            return flask.Response(status=404)

        if "location" in resp.headers:
            return resp

        if "content-type" not in resp.headers:
            resp.headers["content-type"] = data.content_type

        if "content-length" not in resp.headers:
            resp.headers["content-length"] = data.size

        if "content-disposition" not in resp.headers:
            inline_types = config.inline_types()
            disposition = (
                "inline" if send_inline or utils.is_supported_type(data.content_type, inline_types) else "attachment"
            )

            resp.headers.set(
                "content-disposition",
                disposition,
                filename=filename or data.location,
            )

        return resp

    @override
    def temporary_link(self, data: FileData, duration: int, /, **kwargs: Any) -> str:
        try:
            link = super().temporary_link(data, duration, **kwargs)
        except fk.exc.UnsupportedOperationError:
            link = None

        if not link:
            token = utils.encode_token(
                {
                    "topic": "download_file",
                    "exp": str(int(time()) + duration),
                    "storage": self.settings.name,
                    "location": data.location,
                },
            )
            link = tk.url_for("files.temporal_download", token=token, _external=True)
        return link

    @override
    def permanent_link(self, data: FileData, /, **extras: dict[str, Any]) -> str | None:
        """Generate permanent link for the file."""
        if link := super().permanent_link(data, **extras):
            return link

        if self.settings.public:
            return tk.url_for(
                "files.public_download",
                storage_name=self.settings.name,
                location=data.location,
                _external=True,
            )



from __future__ import annotations

from typing import Any

from dominate import tags

import ckan.plugins.toolkit as tk

from ckanext.ap_main.collection.base import (  # type: ignore
    ApCollection,
    ApColumns,
    ApHtmxTableSerializer,
    BulkAction,
    RowAction,
)
from ckanext.collection.types import ButtonFilter, InputFilter
from ckanext.collection.utils import Filters
from ckanext.collection.utils.data.model import ModelData

from ckanext.files.model import File


def file_row_dictizer(serializer: ApHtmxTableSerializer[Any], row: File):
    data = row.dictize({})
    data["bulk-action"] = data["id"]

    return data


class FilesManagerCollection(ApCollection):
    SerializerFactory = ApHtmxTableSerializer.with_attributes(
        record_template="files_manager/record.html",
        row_dictizer=file_row_dictizer,
    )

    ColumnsFactory = ApColumns.with_attributes(
        names=[
            "bulk-action",
            "name",
            "location",
            "size",
            "content_type",
            "storage",
            "created",
            "storage_data",
            "row_actions",
        ],
        sortable={"name", "storage", "content_type", "size", "created"},
        searchable={"name"},
        labels={
            "bulk-action": tk.literal(
                tags.input_(
                    type="checkbox",
                    name="bulk_check",
                    id="bulk_check",
                    data_module="ap-bulk-check",
                    data_module_selector='input[name="entity_id"]',
                ),
            ),
            "name": "Name",
            "location": "Location",
            "storage": "Storage",
            "created": "Uploaded At",
            "storage_data": "Extras",
            "row_actions": "Actions",
        },
        width={"name": "20%", "location": "20%"},
        serializers={
            "storage_data": [("json_display", {})],
        },
    )

    DataFactory = ModelData.with_attributes(
        model=File,
        is_scalar=True,
        use_naive_search=True,
        use_naive_filters=True,
    )

    FiltersFactory = Filters.with_attributes(
        static_actions=[
            BulkAction(
                name="bulk-action",
                type="bulk_action",
                options={
                    "label": "Action",
                    "options": [
                        {"value": "1", "text": "Remove selected files"},
                    ],
                },
            ),
            RowAction(
                name="view",
                type="row_action",
                options={
                    "endpoint": "ap_content.entity_proxy",
                    "label": "View",
                    "params": {
                        "entity_id": "$id",
                        "entity_type": "$type",
                        "view": "read",
                    },
                },
            ),
        ],
        static_filters=[
            InputFilter(
                name="q",
                type="input",
                options={
                    "label": "Search",
                    "placeholder": "Search",
                },
            ),
            ButtonFilter(
                name="type",
                type="button",
                options={
                    "label": "Clear",
                    "type": "button",
                    "attrs": {
                        "onclick": (
                            "$(this).closest('form').find('input,select').val('').prevObject[0].requestSubmit()"
                        ),
                    },
                },
            ),
        ],
    )



"""use timestamp with timezone.

Revision ID: c7081d7f02e8
Revises: d94bacc491e4
Create Date: 2024-06-02 17:55:53.431773

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c7081d7f02e8"
down_revision = "d94bacc491e4"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("files_file", "ctime", type_=sa.DateTime(timezone=True))
    op.alter_column("files_file", "mtime", type_=sa.DateTime(timezone=True))
    op.alter_column("files_file", "atime", type_=sa.DateTime(timezone=True))

    op.alter_column("files_multipart", "ctime", type_=sa.DateTime(timezone=True))


def downgrade():
    op.alter_column("files_file", "ctime", type_=sa.DateTime(timezone=False))
    op.alter_column("files_file", "mtime", type_=sa.DateTime(timezone=False))
    op.alter_column("files_file", "atime", type_=sa.DateTime(timezone=False))

    op.alter_column("files_multipart", "ctime", type_=sa.DateTime(timezone=False))



from __future__ import annotations

import io
import uuid
from typing import Any

import pytest
import sqlalchemy as sa
from faker import Faker
from werkzeug.datastructures import FileStorage

import ckan.model as model
import ckan.plugins.toolkit as tk
from ckan import types
from ckan.tests.factories import fake
from ckan.tests.helpers import call_action  # pyright: ignore[reportUnknownVariableType]

from ckanext.files import shared

call_action: Any


@pytest.fixture(autouse=True)
def prepare(reset_redis: Any):
    reset_redis()


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileCreate:
    def test_unknown_storage(self, file_factory: types.TestFactory, faker: Faker):
        """Unknown storage produces an error."""
        with pytest.raises(tk.ValidationError):
            file_factory(storage=faker.word())

    @pytest.mark.ckan_config("ckanext.files.storage.test.disabled_capabilities", ["CREATE"])
    def test_missing_create_capability(self, file_factory: types.TestFactory):
        """Missing CREATE capability is reported via validation error."""
        with pytest.raises(tk.ValidationError):
            file_factory()

    def test_name_explicit(self, file_factory: types.TestFactory):
        """Name can be overriden even when upload contains filename."""
        name = fake.unique.file_name()
        ignored_name = fake.unique.file_name()
        upload = FileStorage(io.BytesIO(fake.binary(100)), ignored_name)
        result = file_factory(name=name, upload=upload)
        assert result["location"] == name

    def test_missing_name(self, file_factory: types.TestFactory, faker: Faker):
        """If upload does not have filename, explicit name is required."""
        with pytest.raises(tk.ValidationError):
            file_factory(upload=faker.binary(100), name=None)

    def test_name_deduced_from_file(self, file_factory: types.TestFactory):
        """If upload contains filename, explicit name is not required."""
        name = fake.unique.file_name()
        upload = FileStorage(io.BytesIO(fake.binary(100)), name)
        result = file_factory(upload=upload, name=None)
        assert result["location"] == name

    def test_name_secured(self, file_factory: types.TestFactory):
        """Filename is sanitized even without location transformers."""
        bad_name = fake.unique.file_path()
        good_name = bad_name.lstrip("/").replace("/", "_")

        result = file_factory(name=bad_name)
        assert result["location"] == good_name

    @pytest.mark.ckan_config("ckanext.files.storage.test.location_transformers", ["uuid4_prefix"])
    def test_location_transformed(self, file_factory: types.TestFactory):
        """Location transformers are applied to the location."""
        name = fake.unique.file_name()
        result = file_factory(name=name)

        prefix = result["location"][: -len(name)]
        assert uuid.UUID(prefix)

        suffix = result["location"][-len(name) :]
        assert suffix == name

    def test_existing(self, file: dict[str, Any], file_factory: types.TestFactory):
        """Existing file reported via validation error."""
        with pytest.raises(tk.ValidationError):
            file_factory(name=file["location"])

    @pytest.mark.ckan_config("ckanext.files.storage.test.overwrite_existing", True)
    def test_override_does_not_allow_rewriting_file(self, file: dict[str, Any], file_factory: types.TestFactory):
        """Even with enabled overrides, file is not replaced during creation."""
        with pytest.raises(tk.ValidationError):
            file_factory(name=file["location"])

    def test_owner(self, file_factory: types.TestFactory, user: dict[str, Any]):
        """Owner set to the current user."""
        file = file_factory(user="")
        assert file["owner_id"] is None
        assert file["owner_type"] is None

        file = file_factory(user=user)
        assert file["owner_id"] == user["id"]
        assert file["owner_type"] == "user"


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileRegister:
    def test_unknown_storage(self, faker: Faker):
        """Unknown storage produces an error."""
        with pytest.raises(tk.ValidationError):
            call_action("files_file_register", storage=faker.file_name())

    @pytest.mark.ckan_config("ckanext.files.storage.test.disabled_capabilities", ["ANALYZE"])
    def test_missing_analyze_capability(self, faker: Faker):
        """Missing ANALYZE capability is reported via validation error."""
        with pytest.raises(tk.ValidationError):
            call_action("files_file_register", location=faker.file_name())

    def test_already_registered(self, file: dict[str, Any]):
        """If file is registered in the system it causes an error."""
        with pytest.raises(tk.ValidationError):
            call_action("files_file_register", location=file["location"])

    def test_normal_file(self, faker: Faker):
        """Existing file reported via validation error."""
        storage = shared.get_storage()
        info = storage.upload(
            storage.prepare_location(faker.file_name()),
            shared.make_upload(faker.binary(100)),
        )

        result = call_action("files_file_register", location=info.location)
        assert result["location"] == info.location
        assert result["size"] == info.size
        assert result["hash"] == info.hash


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileSearch:
    def test_minimal_search(self, file_factory: types.TestFactory):
        """Search without filters returns all files."""
        first, second = file_factory.create_batch(2)
        result = call_action("files_file_search")
        assert result["count"] == 2
        assert first in result["results"]
        assert second in result["results"]

    def test_sort(self, file_factory: types.TestFactory, faker: Faker):
        """Results can be sorted by multiple fields with different directions."""
        first = file_factory(name="a.txt", upload=faker.binary(10))
        second = file_factory(name="b.txt", upload=faker.binary(5))

        result = call_action("files_file_search", sort="name")
        assert [f["id"] for f in result["results"]] == [first["id"], second["id"]]

        result = call_action("files_file_search", sort="size")
        assert [f["id"] for f in result["results"]] == [second["id"], first["id"]]

        result = call_action("files_file_search", sort=[["size", "desc"]])
        assert [f["id"] for f in result["results"]] == [first["id"], second["id"]]

        result = call_action("files_file_search", sort=[["size", "asc"]])
        assert [f["id"] for f in result["results"]] == [second["id"], first["id"]]

        result = call_action("files_file_search", sort=[["owner_id", "desc"], "size"])
        assert [f["id"] for f in result["results"]] == [second["id"], first["id"]]

        result = call_action("files_file_search", sort=[["owner_id", "asc"], "size"])
        assert [f["id"] for f in result["results"]] == [second["id"], first["id"]]

        result = call_action("files_file_search", sort=["owner_id", ["size", "desc"]])
        assert [f["id"] for f in result["results"]] == [first["id"], second["id"]]

    def test_filter_by_value(self, faker: Faker, file_factory: types.TestFactory):
        """File can be found by model fields."""
        big = file_factory(upload=faker.binary(100))
        small = file_factory(upload=faker.binary(10))

        result = call_action("files_file_search", filters={})
        assert result["count"] == 2

        result = call_action("files_file_search", filters={"size": 10})
        assert result["results"] == [small]

        result = call_action("files_file_search", filters={"size": {"$gte": 50}})
        assert result["results"] == [big]

        result = call_action("files_file_search", filters={"owner_type": {"$eq": "user"}})
        assert big in result["results"]
        assert small in result["results"]

        result = call_action("files_file_search", filters={"location": {"$ne": small["name"]}})
        assert result["results"] == [big]

    def test_filter_by_data(self, file_factory: types.TestFactory):
        """File can be found by storage/plugin data."""
        big_obj = file_factory.model()
        small_obj = file_factory.model()

        big_obj.storage_data = {"big": True, "size": 100, "name": "big"}
        small_obj.storage_data = {"big": False, "size": 10}
        model.Session.commit()

        big = call_action("files_file_show", id=big_obj.id)
        small = call_action("files_file_show", id=small_obj.id)

        result = call_action("files_file_search", filters={"storage_data": {"size": 10}})
        assert result["results"] == [small]

        result = call_action("files_file_search", filters={"storage_data": {"size": {"$gte": 50}}})
        assert result["results"] == [big]

        result = call_action("files_file_search", filters={"storage_data": {"big": {"$ne": None}}})
        assert big in result["results"]
        assert small in result["results"]

        result = call_action("files_file_search", filters={"storage_data": {"big": True}})
        assert result["results"] == [big]

        result = call_action("files_file_search", filters={"storage_data": {"name": None}})
        assert result["results"] == [small]

        result = call_action("files_file_search", filters={"storage_data": {"name": "big"}})
        assert result["results"] == [big]


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileDelete:
    def test_delete_missing(self, faker: Faker):
        """Attempt to remove non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action("files_file_delete", id=faker.uuid4())

    @pytest.mark.ckan_config("ckanext.files.storage.test.disabled_capabilities", ["REMOVE"])
    def test_missing_remove_capability(
        self,
        file: dict[str, Any],
    ):
        """Missing REMOVE capability is reported via validation error."""
        with pytest.raises(tk.ValidationError):
            call_action("files_file_delete", id=file["id"])

    def test_delete_real(self, file: dict[str, Any]):
        """File can be removed."""
        call_action("files_file_delete", id=file["id"])
        existing = model.Session.get(shared.File, file["id"])
        assert not existing


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileShow:
    def test_show_missing(self, faker: Faker):
        """Attempt to show non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action("files_file_show", id=faker.uuid4())

    def test_show_real(self, file: dict[str, Any]):
        """Real file can be shown."""
        result = call_action("files_file_show", id=file["id"])
        assert result == file


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileRename:
    def test_rename_missing(self, faker: Faker):
        """Attempt to rename non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action("files_file_rename", id=faker.uuid4(), name=faker.file_name())

    def test_rename_real(self, faker: Faker, file: dict[str, Any]):
        """Real file can be renamed."""
        name = faker.file_name()
        result = call_action("files_file_rename", id=file["id"], name=name)

        assert result["name"] == name
        assert result["location"] == file["location"]
        assert result["id"] == file["id"]

    def test_name_secured(self, file: dict[str, Any]):
        """Filename is sanitized."""
        bad_name = fake.unique.file_path()
        good_name = bad_name.lstrip("/").replace("/", "_")

        result = call_action("files_file_rename", id=file["id"], name=bad_name)

        assert result["name"] == good_name


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFilePin:
    def test_pin_missing(self, faker: Faker):
        """Attempt to pin non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action("files_file_pin", id=faker.uuid4())

    def test_pin_unowned(self, file_factory: types.TestFactory):
        """Attempt to pin unowned file causes an error."""
        file = file_factory(user="")
        with pytest.raises(tk.ValidationError):
            call_action("files_file_pin", id=file["id"])

    def test_pin_real(self, file: dict[str, Any]):
        """Real file can be pinned."""
        result = call_action("files_file_pin", id=file["id"])
        assert result["pinned"]


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileUnpin:
    def test_unpin_missing(self, faker: Faker):
        """Attempt to unpin non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action("files_file_unpin", id=faker.uuid4())

    def test_unpin_unowned(self, file_factory: types.TestFactory):
        """Unowned file can be unpinned(nothing is changed)."""
        file = file_factory(user="")
        result = call_action("files_file_unpin", id=file["id"])
        assert not result["pinned"]

    def test_unpin_real(self, file: dict[str, Any]):
        """Real file can be unpinned."""
        result = call_action("files_file_unpin", id=file["id"])
        assert not result["pinned"]


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileOwnershipTransfer:
    def test_transfer_missing(self, faker: Faker):
        """Attempt to transfer non-existing file causes an error."""
        with pytest.raises(tk.NotFound):
            call_action(
                "files_transfer_ownership",
                id=faker.uuid4(),
                owner_id=fake.unique.uuid4(),
                owner_type="user",
            )

    def test_transfer_pinned(self, faker: Faker, file: dict[str, Any]):
        """Attempt to transfer pinned file causes an error without force-flag."""
        call_action("files_file_pin", id=file["id"])
        new_owner = faker.uuid4()
        with pytest.raises(tk.ValidationError):
            call_action(
                "files_transfer_ownership",
                id=file["id"],
                owner_id=new_owner,
                owner_type="user",
            )

        result = call_action(
            "files_transfer_ownership",
            id=file["id"],
            owner_id=new_owner,
            owner_type="user",
            force=True,
        )
        assert result["owner_id"] == new_owner

    def test_transfer_history(self, file_factory: types.TestFactory[shared.File]):
        """Transfer history is recorded when owner changes."""
        fileobj = file_factory.model()

        assert fileobj.owner
        stmt = fileobj.owner.select_history().with_only_columns(sa.func.count())
        size = model.Session.scalar(stmt)
        assert not size

        # owner remains the same, history is not recorded
        call_action(
            "files_transfer_ownership",
            id=fileobj.id,
            owner_id=fileobj.owner.owner_id,
            owner_type=fileobj.owner.owner_type,
        )
        size = model.Session.scalar(stmt)
        assert not size

        call_action(
            "files_transfer_ownership",
            id=fileobj.id,
            owner_id=fake.unique.uuid4(),
            owner_type=fileobj.owner.owner_type,
        )
        size = model.Session.scalar(stmt)
        assert size == 1

    def test_transfer_unowned(self, faker: Faker, file_factory: types.TestFactory[shared.File]):
        """Unowned file can be transfered without additional conditions."""
        file = file_factory.model(user="")
        assert not file.owner

        owner_id = fake.unique.uuid4()
        owner_type = faker.word()

        result = call_action(
            "files_transfer_ownership",
            id=file.id,
            owner_id=owner_id,
            owner_type=owner_type,
        )

        assert result["owner_id"] == owner_id
        assert result["owner_type"] == owner_type

    def test_transfer_with_pin(self, faker: Faker, file: dict[str, Any]):
        """File can be pinned during transfer."""
        result = call_action(
            "files_transfer_ownership",
            id=file["id"],
            owner_id=fake.unique.uuid4(),
            owner_type=faker.word(),
        )
        assert not result["pinned"]

        result = call_action(
            "files_transfer_ownership",
            id=file["id"],
            owner_id=fake.unique.uuid4(),
            owner_type=faker.word(),
            pin=True,
        )
        assert result["pinned"]

    def test_transfer_expires_owner(self, faker: Faker, file_factory: types.TestFactory):
        """Owner attribute of the file is refreshed during transfer."""
        file = file_factory.model(user="")
        assert not file.owner

        call_action(
            "files_transfer_ownership",
            id=file.id,
            owner_id=faker.uuid4(),
            owner_type="user",
        )
        assert file.owner


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileOwnerScan:
    def test_filter_by_owner(self, file_factory: types.TestFactory, user: dict[str, Any]):
        """Files are filtered by owner."""
        owned = file_factory(user=user["name"])
        file_factory(user="")

        result = call_action("files_file_scan", owner_type="user", owner_id=user["id"])
        assert result["results"] == [owned]

        result = call_action("files_file_scan", owner_type="user", owner_id=fake.unique.uuid4())
        assert result["results"] == []



from __future__ import annotations

import pydoc
import textwrap

import click

from ckan.config.declaration import Declaration, Key

from ckanext.files import base

from . import dev, file, maintain, migrate, stats, storage

try:
    import ckan.cli.files  # type: ignore # noqa

    entrypoint = "ex-files"
except ImportError:
    entrypoint = "files"


__all__ = [
    "files",
]


@click.group(entrypoint, short_help="ckanext-files CLI commands")
def files():
    pass


files.add_command(dev.group, "dev")
files.add_command(migrate.group, "migrate")
files.add_command(stats.group, "stats")
files.add_command(storage.group, "storage")
files.add_command(maintain.group, "maintain")
files.add_command(file.group, "file")


@files.command()
@click.option("-c", "--with-configuration", is_flag=True, help="Show adapter's configuration")
@click.option("-d", "--with-docs", is_flag=True, help="Show adapter's documentation")
@click.option("-H", "--include-hidden", is_flag=True, help="Show hidden adapters")
@click.argument("adapter", required=False)
def adapters(
    adapter: str | None,
    with_docs: bool,
    include_hidden: bool,
    with_configuration: bool,
):
    """Show all awailable storage adapters."""
    for name in sorted(base.adapters):
        if adapter and name != adapter:
            continue

        item = base.adapters[name]
        if item.hidden and not include_hidden:
            continue

        click.secho(
            f"{click.style(name, bold=True)} - {item.__module__}:{item.__name__}",
        )

        if with_docs and (doc := pydoc.getdoc(item)):
            doc = f"{click.style('Documentation:', bold=True)}\n{doc}"
            wrapped = textwrap.indent(doc, "\t")
            click.secho(wrapped)
            click.echo()

        if with_configuration and issubclass(item, base.Storage):
            decl = Declaration()
            item.declare_config_options(decl, Key.from_string("ckanext.files.storage.NAME"))
            configuration = decl.into_ini(False, True)
            configuration = f"{click.style('Configuration:', bold=True)}\n{configuration}"
            wrapped = textwrap.indent(configuration, "\t")
            click.secho(wrapped)
            click.echo()



"""empty message.

Revision ID: ea61ab6a6042
Revises: 6e3c2aa192f6
Create Date: 2026-04-03 16:11:40.353869

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "ea61ab6a6042"
down_revision = "6e3c2aa192f6"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("files_file", sa.Column("algorithm", sa.Text(), nullable=False, server_default=""))
    op.alter_column("files_file", "ctime", new_column_name="created")
    op.drop_column("files_file", "mtime")
    op.drop_column("files_file", "atime")


def downgrade():
    op.add_column(
        "files_file", sa.Column("atime", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True)
    )
    op.add_column(
        "files_file", sa.Column("mtime", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True)
    )
    op.alter_column("files_file", "created", new_column_name="ctime")
    op.drop_column("files_file", "algorithm")



"""Types for the extension."""

from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Literal, Protocol

from file_keeper.core.types import LocationTransformer

from ckan.config.declaration import Declaration, Key
from ckan.types import (
    Context,
    FlattenDataDict,
    FlattenErrorDict,
    FlattenKey,
    Response,
    Validator,
    ValidatorFactory,
)

FileOperation = Literal["show", "update", "delete"]
"""Operation that performed on file."""

OwnerOperation = Literal["file_transfer", "file_scan"]
"""Operation that performed on owner."""


class PUploadStream(Protocol):
    def read(self, size: Any = ..., /) -> bytes: ...

    def __iter__(self) -> Iterator[bytes]: ...


class PTask(Protocol):
    def __call__(self, result: Any, idx: int, prev: Any) -> Any: ...


__all__ = [
    "Context",
    "Validator",
    "ValidatorFactory",
    "FlattenKey",
    "FlattenErrorDict",
    "FlattenDataDict",
    "Declaration",
    "FileOperation",
    "OwnerOperation",
    "Key",
    "PTask",
    "Response",
    "LocationTransformer",
]



from __future__ import annotations

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, relationship

from ckan.lib.dictization import table_dictize
from ckan.model.vocabulary import TYPE_CHECKING
from ckan.types import Context

if TYPE_CHECKING:
    from .file import FilesFile


from .base import Base


class FilesOwner(Base):
    """Model with details about current owner of an item.

    Keyword Args:
        file_id (str): ID of the owned object
        owner_id (str): ID of the owner
        owner_type (str): Type of the owner
        pinned (bool): is ownership protected from transfer

    Example:
        ```python
        owner = Owner(
            file_id=file.id,
            owner_id=user.id,
            owner_type="user,
        )
        ```
    """

    __table__ = sa.Table(
        "files_owner",
        Base.metadata,
        sa.Column("file_id", sa.Text, primary_key=True),
        sa.Column("owner_id", sa.Text, nullable=False),
        sa.Column("owner_type", sa.Text, nullable=False),
        sa.Column("pinned", sa.Boolean, default=False, nullable=False),
        sa.Index("idx_file_owner_owner", "owner_type", "owner_id", unique=False),
        sa.ForeignKeyConstraint(
            ["file_id"],
            ["files_file.id"],
            "files_owner_file_id_fkey",
            ondelete="CASCADE",
        ),
    )
    file_id: Mapped[str]
    owner_id: Mapped[str]
    owner_type: Mapped[str]
    pinned: Mapped[bool]

    files: Mapped[list[FilesFile]] = relationship(back_populates="owner")

    def dictize(self, context: Context):
        return table_dictize(self, context)

    def select_history(self):
        """Returns a select statement to fetch ownership history."""
        from .transfer_history import TransferHistory  # noqa: PLC0415

        return (
            sa.select(TransferHistory)
            .join(FilesOwner)
            .where(
                TransferHistory.file_id == self.file_id,
            )
        )



from __future__ import annotations

from typing import Any, cast

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import authz, model
from ckan.types import AuthResult, Context

from ckanext.files import interfaces, shared, types, utils


def _owner_allows(
    context: Context,
    owner_type: str,
    owner_id: str,
    operation: types.OwnerOperation,
) -> bool:
    """Decide if user is allowed to perform operation on owner."""
    for plugin in p.PluginImplementations(interfaces.IFiles):
        result = plugin.files_owner_allows(context, owner_type, owner_id, operation)
        if result is not None:
            return result

    if (operation == "file_transfer" and shared.config.transfer_as_update()) or (
        operation == "file_scan" and shared.config.scan_as_update()
    ):
        func_name = f"{owner_type}_update"

    else:
        func_name = f"{owner_type}_{operation}"

    result = authz.is_authorized(func_name, tk.fresh_context(context), {"id": owner_id})
    return result["success"]


def _file_allows(
    context: Context,
    file: shared.File,
    operation: types.FileOperation,
) -> bool:
    """Decide if user is allowed to perform operation on file."""
    for plugin in p.PluginImplementations(interfaces.IFiles):
        result = plugin.files_file_allows(context, file, operation)
        if result is not None:
            return result

    info = file.owner if file else None

    if not info:
        return False

    cascade = shared.config.cascade_access()
    if info.owner_type not in cascade:
        return False

    if cascade[info.owner_type] and file.storage not in cascade[info.owner_type]:
        return False

    func_name = f"{info.owner_type}_{operation}"

    result = authz.is_authorized(
        func_name,
        tk.fresh_context(context),
        {"id": info.owner_id},
    )

    return result["success"]


def _get_user(context: Context) -> model.User | None:
    user = context.get("auth_user_obj")
    if isinstance(user, model.User):
        return user

    if tk.current_user and tk.current_user.name == context["user"]:
        return cast(model.User, tk.current_user)

    cache = utils.ContextCache(context)
    return cache.get("user", context["user"], lambda: model.User.get(context["user"]))


def _get_file(context: Context, file_id: str) -> shared.File | None:
    cache = utils.ContextCache(context)
    return cache.get_model("file", file_id, shared.File)


# Permissions #################################################################


@tk.auth_allow_anonymous_access
def files_permission_manage_files(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to manage any file.

    This is a sort of "sysadmin" check in terms of file management. Give this
    permission to user who needs an access to every owned, unowned, hidden,
    incomplete and private file
    """
    return {"success": False, "msg": "Not allowed to manage files"}


@tk.auth_allow_anonymous_access
def files_permission_owns_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to manage a file.

    Normally, owner of the file passes this check as well as any user who has
    ``files_permission_manage_files``.
    """
    if authz.is_authorized_boolean("files_permission_manage_files", context, data_dict):
        return {"success": True}

    not_an_owner = "Not an owner of the file"
    user = _get_user(context)
    if not user:
        return {"success": False, "msg": not_an_owner}

    file = _get_file(context, data_dict["id"])
    if not file or not file.owner:
        return {"success": False, "msg": not_an_owner}

    return {
        "success": file.owner.owner_type == "user" and file.owner.owner_id == user.id,
        "msg": not_an_owner,
    }


@tk.auth_allow_anonymous_access
def files_permission_edit_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to edit a file.

    Owners and global managers can edit files. Additionally plugins can extend
    this permission via ``IFiles.files_file_allows`` hook.

    """
    result = authz.is_authorized_boolean("files_permission_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"])
        result = bool(file and _file_allows(context, file, "update"))

    return {"success": result, "msg": "Not allowed to edit file"}


@tk.auth_allow_anonymous_access
def files_permission_delete_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to delete the file.

    Owners and global managers can delete files. Additionally plugins can extend
    this permission via ``IFiles.files_file_allows`` hook.

    """
    result = authz.is_authorized_boolean("files_permission_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"])
        result = bool(file and _file_allows(context, file, "delete"))

    return {"success": result, "msg": "Not allowed to delete file"}


@tk.auth_allow_anonymous_access
def files_permission_read_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to read a file.

    Owners and global managers can read files. Additionally plugins can extend
    this permission via ``IFiles.files_file_allows`` hook.

    """
    result = authz.is_authorized_boolean("files_permission_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"])
        result = bool(file and _file_allows(context, file, "show"))

    return {"success": result, "msg": "Not allowed to read file"}


@tk.auth_allow_anonymous_access
def files_permission_download_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user is allowed to download a file.

    Owners and global managers can download files. Additionally plugins can extend
    this permission via ``IFiles.files_file_allows`` hook.

    """
    result = authz.is_authorized_boolean("files_permission_read_file", context, data_dict)
    return {"success": result, "msg": "Not allowed to read file"}


# API #########################################################################


@tk.auth_allow_anonymous_access
def files_file_search(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file manager can search files."""
    return authz.is_authorized("files_permission_manage_files", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_create(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    if context["user"] and (
        shared.config.authenticated_uploads() and data_dict["storage"] in shared.config.authenticated_storages()
    ):
        return {"success": True}

    return authz.is_authorized("files_permission_manage_files", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_register(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Check if user can register files from storage in DB.

    Only file manager can register files.
    """
    return authz.is_authorized("files_permission_manage_files", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_delete(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can remove files."""
    return authz.is_authorized("files_permission_delete_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_show(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can view files."""
    return authz.is_authorized("files_permission_read_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_rename(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can rename files."""
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_pin(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_unpin(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_transfer_ownership(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file manager can transfer ownership."""
    file = _get_file(context, data_dict["id"])
    if not file or (file.owner and file.owner.pinned and not data_dict["force"]):
        return {"success": False, "msg": "File is pinned"}

    result = authz.is_authorized_boolean("files_permission_manage_files", context, data_dict)
    if not result:
        result = bool(
            authz.is_authorized_boolean("files_permission_edit_file", context, data_dict)
            and _owner_allows(
                context,
                data_dict["owner_type"],
                data_dict["owner_id"],
                "file_transfer",
            ),
        )

    return {"success": result, "msg": "Not allowed to edit file"}


@tk.auth_allow_anonymous_access
def files_file_scan(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file owner can list files."""
    result = authz.is_authorized_boolean("files_permission_manage_files", context, data_dict)
    if not result:
        result = _owner_allows(
            context,
            data_dict["owner_type"],
            data_dict["owner_id"],
            "file_scan",
        )

    return {"success": result, "msg": "Not allowed to list files"}


# not included into CKAN ######################################################


@tk.auth_allow_anonymous_access
def files_resource_upload(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Users can upload resources files."""
    # TODO: add restriction on number of free-files or their total size inside
    # resource storage.
    try:
        shared.get_storage(shared.config.resources_storage())
    except shared.exc.UnknownStorageError:
        return {"success": False}

    if data_dict.get("resource_id"):
        return authz.is_authorized("resource_update", context, {"id": data_dict["resource_id"]})

    return authz.is_authorized("resource_create", context, data_dict)


@tk.auth_allow_anonymous_access
def files_autocomplete_available_resource_files(
    context: Context,
    data_dict: dict[str, Any],
) -> AuthResult:
    """Users allowed to search their free resource files."""
    return {"success": True}


@tk.auth_allow_anonymous_access
def files_file_replace(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can replace files."""
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_multipart_refresh(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_multipart_start(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_file_create", context, data_dict)


@tk.auth_allow_anonymous_access
def files_multipart_update(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_permission_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_multipart_complete(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized("files_permission_edit_file", context, data_dict)



from file_keeper import Capability, HashingReader, Location, Upload, make_upload
from file_keeper.core import exceptions as exc

from . import config, types
from .base import (
    FileData,
    Manager,
    Reader,
    Settings,
    Storage,
    Uploader,
    get_storage,
    make_storage,
)
from .interfaces import IFiles
from .model import File, Owner, TransferHistory
from .task import Task, TaskQueue, add_task, with_task_queue

__all__ = [
    "get_storage",
    "make_storage",
    "make_upload",
    "Upload",
    "HashingReader",
    "Capability",
    "File",
    "Location",
    "Owner",
    "TransferHistory",
    "FileData",
    "IFiles",
    "Storage",
    "Settings",
    "Uploader",
    "Reader",
    "Manager",
    "add_task",
    "with_task_queue",
    "Task",
    "TaskQueue",
    "types",
    "config",
    "exc",
]



from __future__ import annotations

import dataclasses

from file_keeper.default.adapters import sqlalchemy
from typing_extensions import override

from ckanext.files import shared, types


@dataclasses.dataclass()
class Settings(shared.Settings, sqlalchemy.Settings):
    pass


class DbStorage(shared.Storage, sqlalchemy.SqlAlchemyStorage):  # pyright: ignore[reportIncompatibleVariableOverride]
    hidden = True

    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = Settings
    ReaderFactory = type("Reader", (shared.Reader, sqlalchemy.Reader), {})
    ManagerFactory = type("Manager", (shared.Manager, sqlalchemy.Manager), {})
    UploaderFactory = type("Uploader", (shared.Uploader, sqlalchemy.Uploader), {})

    @override
    @classmethod
    def declare_config_options(
        cls,
        declaration: types.Declaration,
        key: types.Key,
    ):
        declaration.declare(key.db_url).required()
        declaration.declare(key.table_name).required()
        declaration.declare(key.location_column).required()
        declaration.declare(key.content_column).required()



from __future__ import annotations

import dataclasses
from typing import Any

from file_keeper.default.adapters import s3
from typing_extensions import override

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.config.declaration import Declaration, Key

from ckanext.files import shared


class Reader(shared.Reader, s3.Reader):
    @override
    def response(self, data: shared.FileData, extras: dict[str, Any]) -> types.Response:
        return tk.redirect_to(self.temporary_link(data, 60, extras))


@dataclasses.dataclass()
class Settings(shared.Settings, s3.Settings):
    pass


class S3Storage(shared.Storage, s3.S3Storage):  # pyright: ignore[reportIncompatibleVariableOverride]
    """AWS S3 adapter."""

    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = Settings
    ReaderFactory = type("Reader", (shared.Reader, s3.Reader), {})
    UploaderFactory = type("Reader", (shared.Uploader, s3.Uploader), {})
    ManagerFactory = type("Reader", (shared.Manager, s3.Manager), {})

    @override
    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)

        declaration.declare(key.bucket).required().set_description(
            "Name of the S3 bucket where uploaded data will be stored.",
        )
        declaration.declare(key.key).set_description("The access key for AWS account.")
        declaration.declare(key.secret).set_description("The secret key for AWS account.")
        declaration.declare(key.endpoint).set_description("Custom AWS endpoint.")

        declaration.declare(key.region).set_description("The AWS Region used in instantiating the client.")



from __future__ import annotations

from datetime import datetime, timezone

import click
import file_keeper as fk
import sqlalchemy as sa
from babel.dates import format_datetime, format_timedelta

import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.files import shared


def _now():
    return datetime.now(timezone.utc)


@click.group()
def group():
    """Storage statistics."""


storage_option = click.option(
    "-s",
    "--storage-name",
    help="Name of the configured storage",
)


@group.command()
@storage_option
def overview(storage_name: str | None):
    """General information about storage usage."""
    storage_name = storage_name or shared.config.default_storage()
    try:
        shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError as err:
        tk.error_shout(f"Storage {storage_name} is not configured")
        raise click.Abort from err

    stmt = sa.select(
        sa.func.sum(shared.File.size),
        sa.func.count(shared.File.id),
        sa.func.max(shared.File.created),
        sa.func.min(shared.File.created),
    ).where(shared.File.storage == storage_name)
    row = model.Session.execute(stmt).fetchone()
    size, count, newest, oldest = row if row else (0, 0, _now(), _now())

    if not count:
        tk.error_shout(f"Storage {storage_name} is empty")
        raise click.Abort

    click.secho(f"Number of files: {click.style(count, bold=True)}")
    click.secho(
        f"Used space: {click.style(fk.humanize_filesize(size), bold=True)}",
    )
    click.secho(
        "Newest file created at: "
        + f"{click.style(format_datetime(newest), bold=True)} "
        + f"({format_timedelta(newest - _now(), add_direction=True)})",
    )
    click.secho(
        "Oldest file created at: "
        + f"{click.style(format_datetime(oldest), bold=True)} "
        + f"({format_timedelta(oldest - _now(), add_direction=True)})",
    )


@group.command()
@storage_option
def types(storage_name: str | None):
    """Files distribution by MIMEtype."""
    storage_name = storage_name or shared.config.default_storage()
    stmt = (
        sa.select(
            shared.File.content_type,
            sa.func.count(shared.File.content_type).label("count"),
        )
        .where(shared.File.storage == storage_name)
        .group_by(shared.File.content_type)
        .order_by(shared.File.content_type)
    )

    total = model.Session.scalar(sa.select(sa.func.sum(stmt.c.count)))
    click.secho(
        f"Storage {click.style(storage_name, bold=True)} contains " + f"{click.style(total, bold=True)} files",
    )
    for content_type, count in model.Session.execute(stmt):
        click.secho(f"\t{content_type}: {click.style(count, bold=True)}")


@group.command()
@storage_option
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Show distribution for every owner ID",
)
def owner(storage_name: str | None, verbose: bool):
    """Files distribution by owner."""
    storage_name = storage_name or shared.config.default_storage()
    owner_col = (
        sa.func.concat(shared.Owner.owner_type, " ", shared.Owner.owner_id)
        if verbose
        else sa.func.concat(shared.Owner.owner_type, "")
    )

    stmt = (
        sa.select(
            owner_col.label("owner"),
            sa.func.count(shared.File.id),
        )
        .where(shared.File.storage == storage_name)
        .outerjoin(
            shared.Owner,
            sa.and_(
                shared.Owner.file_id == shared.File.id,
            ),
        )
        .group_by(owner_col)
    ).order_by(owner_col)

    total = model.Session.scalar(sa.select(sa.func.sum(stmt.c.count)))
    click.secho(
        f"Storage {click.style(storage_name, bold=True)} contains " + f"{click.style(total, bold=True)} files",
    )
    for owner, count in model.Session.execute(stmt):
        clean_owner = owner.strip() or click.style(
            "has no owner",
            underline=True,
            bold=True,
        )
        click.secho(
            f"\t{clean_owner}: {click.style(count, bold=True)}",
        )



from __future__ import annotations

import logging
from functools import partial
from typing import Any

import file_keeper as fk
import jwt
from flask import Blueprint, jsonify

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.common import streaming_response
from ckan.lib.pagination import Page
from ckan.logic import parse_params
from ckan.types import Response
from ckan.views.resource import download

from ckanext.files import shared, utils

log = logging.getLogger(__name__)
bp = Blueprint("files", __name__)

__all__ = ["bp"]


def not_found_handler(error: tk.ObjectNotFound) -> tuple[str, int]:
    """Generic handler for ObjectNotFound exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 404,
                "content": f"Object not found: {error.message}",
                "name": "Not found",
            },
        ),
        404,
    )


def not_authorized_handler(error: tk.NotAuthorized) -> tuple[str, int]:
    """Generic handler for NotAuthorized exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 403,
                "content": error.message or "Not authorized to view this page",
                "name": "Not authorized",
            },
        ),
        403,
    )


bp.register_error_handler(tk.ObjectNotFound, not_found_handler)
bp.register_error_handler(tk.NotAuthorized, not_authorized_handler)


def get_blueprints():
    return [bp]


def _as_response(storage_name: str, data: shared.FileData):
    """Return a response for a file stored in the given location.

    The storage is looked up by name, and the file data is created from the
    location string. This is a helper function for the all kind of download
    routes.
    """
    try:
        storage = shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError:
        return tk.abort(404)

    if isinstance(storage, shared.Storage):
        resp = storage.as_response(data)
        if resp.status_code >= 400:
            return tk.abort(resp.status_code)
        return resp

    return tk.abort(422, "File is not downloadable")


@bp.route("/files/download/<file_id>")
def dispatch_download(file_id: str) -> Response:
    tk.check_access("files_permission_download_file", {}, {"id": file_id})
    item: dict[str, Any] = tk.get_action("file_show")({}, {"id": file_id})

    return _as_response(item["storage"], shared.FileData.from_dict(item))


@bp.route("/files/public-download/<storage_name>/<path:location>")
def public_download(storage_name: str, location: str) -> Response:
    """Download a public file by its storage name and location.

    The file is looked up by its storage name and location, and the content is
    returned if the storage is public. This route is intended for files that
    are not tracked in DB, but are stored in a public storage. The user does
    not need any permissions to access the file, but the storage must be
    explicitly marked as public to prevent unauthorized access to private
    files. The file content is returned as a response, using the same mechanism
    as the normal download route.

    If the storage is not found or not public, a 404 or 403 error is returned
    respectively. If the file is not found in the storage, a 404 error is
    returned.
    """
    try:
        storage = shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError:
        return tk.abort(404)

    if not isinstance(storage, shared.Storage) or not storage.settings.public:
        return tk.abort(403, "Storage is not public")

    location = shared.Location(location)
    try:
        data = shared.FileData(
            location,
            size=storage.size(location),
            content_type=storage.content_type(location),
        )
    except shared.exc.MissingFileError:
        return tk.abort(404)

    return _as_response(storage_name, data)


def _streaming_file(
    item: shared.File,
    storage: fk.Storage,
    data: shared.FileData,
) -> Response | None:
    if storage.supports(shared.Capability.STREAM):
        resp = streaming_response(storage.stream(data), data.content_type)
        if utils.is_supported_type(item.content_type, shared.config.inline_types()):
            resp.headers["content-disposition"] = f"inline; filename={item.name}"
        else:
            resp.headers["content-disposition"] = f"attachment; filename={item.name}"

        model.Session.commit()

        return resp


@bp.route("/files/token-download/<token>")
def temporal_download(token: str) -> Response:
    try:
        data = utils.decode_token(token)

    except jwt.ExpiredSignatureError as err:
        log.debug("Expired file-download token: %s", err)
        raise tk.ObjectNotFound("file") from err

    except jwt.InvalidTokenError as err:
        log.debug("Cannot decode file-download token: %s", err)
        raise tk.ObjectNotFound("file") from err

    if data.get("topic") != "download_file":
        raise tk.ObjectNotFound("file")

    if "id" in data:
        item = model.Session.get(shared.File, data["sub"])
    elif "location" in data and "storage" in data:
        item = model.Session.scalar(shared.File.by_location(data["location"], data["storage"]))
    else:
        item = None

    if not item:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(item.storage)
    data = shared.FileData.from_object(item)

    if isinstance(storage, shared.Storage):
        return storage.as_response(data)

    if resp := _streaming_file(item, storage, data):
        return resp

    return tk.abort(422, "File is not downloadable")


def _pager_url(*args: Any, **kwargs: Any) -> str:
    """Generic URL builder for `url` parameter of ckan.lib.pagination.Page.

    It generates pagination link keeping all the parameters from URL and query
    string.
    """
    params = {k: v for k, v in tk.request.args.items() if k != "page"}
    view_args: dict[str, Any] = tk.request.view_args or {}
    params.update(
        dict(view_args.items()),
    )
    params.update(kwargs)
    return tk.h.pager_url(*args, **params)


@bp.route("/<owner_type>/<owner_id>/files/list")
def list_files(owner_type: str, owner_id: str) -> str:
    owner = utils.get_owner(owner_type, owner_id)
    if owner and hasattr(owner, "id"):
        owner_id = owner.id

    rows = 10
    params = tk.request.args
    page = tk.h.get_page_number(params)
    start = rows * page - rows

    search_dict: dict[str, Any] = {
        "rows": rows,
        "start": start,
        "owner_type": owner_type,
        "owner_id": owner_id,
        "sort": params.get("sort", "created"),
        "reverse": params.get("reverse", True),
    }

    if "storage" in params:
        search_dict["storage"] = params["storage"]

    result: dict[str, Any]
    try:
        result = tk.get_action("files_file_scan")({}, search_dict)
    except tk.ValidationError as err:
        for k, v in err.error_summary.items():
            tk.h.flash_error(f"{k}: {v}")

        result = {"count": 0, "results": []}

    pager = Page(
        result["results"],
        items_per_page=rows,
        page=page,
        item_count=result["count"],
        url=partial(_pager_url),
    )

    tpl_data: dict[str, Any] = {
        "owner": owner,
        "pager": pager,
        "owner_type": owner_type,
        "owner_id": owner_id,
    }

    return tk.render("files/index.html", tpl_data)


@bp.route("/<owner_type>/<owner_id>/files/delete/<file_id>", methods=["POST", "GET"])
def delete_file(
    owner_type: str,
    owner_id: str,
    file_id: str,
) -> str | shared.types.Response:
    if tk.request.method == "POST":
        try:
            tk.get_action("files_file_delete")({}, {"id": file_id})
        except tk.NotAuthorized as err:
            tk.h.flash_error(err)
        else:
            came_from = tk.h.get_request_param("came_from")
            if came_from:
                return tk.redirect_to(came_from)

            return tk.redirect_to(
                "files.list_files",
                owner_type=owner_type,
                owner_id=owner_id,
            )

    tk.check_access("files_file_delete", {}, {"id": file_id})
    info = tk.get_action("files_file_show")({}, {"id": file_id})
    return tk.render("files/delete.html", {"file": info})


@bp.route("/api/util/files_autocomplete_own_files")
def autocomplete_own_files() -> Any:
    tk.check_access("files_autocomplete_available_resource_files", {}, {})

    params: dict[str, Any] = parse_params(tk.request.args)

    q = params.pop("q", "")

    params.update(
        {
            "owner_type": "user",
            "owner_id": tk.current_user.is_authenticated and tk.current_user.id,  # type: ignore
            "pinned": False,
            "name": ["like", f"%{q}%"],
        },
    )

    result = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        params,
    )

    return jsonify(
        {
            "ResultSet": {
                "Result": [
                    dict(
                        item,
                        label="{name} [{content_type}, {size}]".format(
                            name=item["name"],
                            content_type=tk.h.unified_resource_format(
                                item["content_type"],
                            ),
                            size=fk.humanize_filesize(item["size"]),
                        ),
                    )
                    for item in result["results"]
                ],
            },
        },
    )


@bp.route("/api/util/files_autocomplete_available_resource_files")
def autocomplete_available_resource_files() -> Any:
    tk.check_access("files_autocomplete_available_resource_files", {}, {})
    q = tk.request.args.get("incomplete")

    result = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {
            "owner_type": "user",
            "owner_id": tk.current_user.is_authenticated and tk.current_user.id,
            "pinned": False,
            "storage": shared.config.resources_storage(),
            "name": ["like", f"%{q}%"],
        },
    )

    return jsonify(
        {
            "ResultSet": {
                "Result": [
                    {
                        "id": item["id"],
                        "name": item["name"],
                        "label": "{name} [{content_type}, {size}]".format(
                            name=item["name"],
                            content_type=tk.h.unified_resource_format(
                                item["content_type"],
                            ),
                            size=fk.humanize_filesize(item["size"]),
                        ),
                    }
                    for item in result["results"]
                ],
            },
        },
    )


@bp.route(
    "/<package_type>/<id>/resource/<resource_id>/download",
    defaults={"package_type": "dataset"},
)
@bp.route(
    "/<package_type>/<id>/<resource_id>/download/<filename>",
    defaults={"package_type": "dataset"},
)
def resource_download(
    package_type: str,
    id: str,
    resource_id: str,
    filename: str | None = None,
):
    try:
        resource = tk.get_action("resource_show")({}, {"id": resource_id})
    except tk.ObjectNotFound:
        return tk.abort(404, tk._("Resource not found"))
    except tk.NotAuthorized:
        return tk.abort(403, tk._("Not authorized to download resource"))

    if resource.get("url_type") != "file":
        return download(package_type, id, resource_id, filename)

    file_id = resource["url"].rsplit("/", 1)[-1]
    return dispatch_download(file_id)



from ckan.logic.schema import validator_args
from ckan.types import Schema, Validator, ValidatorFactory

from ckanext.files import shared


@validator_args
def file_create(  # noqa: PLR0913
    ignore_empty: Validator,
    unicode_safe: Validator,
    default: ValidatorFactory,
    files_into_upload: Validator,
    not_missing: Validator,
) -> Schema:
    # name is checked inside action, using "upload" as source if empty
    return {
        "name": [ignore_empty, unicode_safe],
        "storage": [default(shared.config.default_storage()), unicode_safe],
        "upload": [not_missing, files_into_upload],
    }


@validator_args
def file_register(
    default: ValidatorFactory,
    unicode_only: Validator,
    not_missing: Validator,
) -> Schema:
    return {
        "location": [not_missing, unicode_only],
        "storage": [
            default(shared.config.default_storage()),
            unicode_only,
        ],
    }


@validator_args
def file_search(
    default: ValidatorFactory,
    int_validator: Validator,
    dict_only: Validator,
    convert_to_json_if_string: Validator,
) -> Schema:
    return {
        "start": [default(0), int_validator],
        "rows": [default(10), int_validator],
        "sort": [default("name")],
        "filters": [default("{}"), convert_to_json_if_string, dict_only],
    }


@validator_args
def file_delete(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe]}


@validator_args
def file_show(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe]}


@validator_args
def file_rename(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe], "name": [not_empty, unicode_safe]}


@validator_args
def transfer_ownership(
    not_empty: Validator,
    boolean_validator: Validator,
    default: ValidatorFactory,
    unicode_safe: Validator,
) -> Schema:
    return {
        "id": [not_empty, unicode_safe],
        "owner_id": [not_empty, unicode_safe],
        "owner_type": [not_empty, unicode_safe],
        "force": [default(False), boolean_validator],
        "pin": [default(False), boolean_validator],
    }


@validator_args
def file_scan(default: ValidatorFactory, unicode_only: Validator, ignore_missing: Validator) -> Schema:
    return {
        "owner_id": [default(""), unicode_only],
        "owner_type": [default("user"), unicode_only],
        "start": [ignore_missing],
        "rows": [ignore_missing],
        "sort": [ignore_missing],
    }


@validator_args
def file_pin(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe]}


@validator_args
def file_unpin(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe]}


# not included into CKAN ######################################################


@validator_args
def resource_upload(
    keep_extras: Validator,
    unicode_safe: Validator,
    boolean_validator: Validator,
    ignore_missing: Validator,
) -> Schema:
    return {
        "multipart": [boolean_validator],
        "resource_id": [ignore_missing, unicode_safe],
        "package_id": [ignore_missing, unicode_safe],
        "__extras": [keep_extras],
    }


@validator_args
def group_image_upload(ignore: Validator) -> Schema:
    schema = file_create()
    schema["storage"] = [ignore]
    schema["__extras"] = [ignore]
    return schema


@validator_args
def user_image_upload(ignore: Validator) -> Schema:
    schema = file_create()
    schema["storage"] = [ignore]
    schema["__extras"] = [ignore]
    return schema


@validator_args
def file_replace(  # noqa: PLR0913
    unicode_safe: Validator,
    files_into_upload: Validator,
    not_missing: Validator,
) -> Schema:
    return {
        "id": [not_missing, unicode_safe],
        "upload": [not_missing, files_into_upload],
    }


@validator_args
def multipart_start(
    not_empty: Validator,
    unicode_safe: Validator,
    default: ValidatorFactory,
    int_validator: Validator,
    ignore_missing: Validator,
    files_into_upload: Validator,
) -> Schema:
    return {
        "storage": [default(shared.config.default_storage()), unicode_safe],
        "name": [not_empty, unicode_safe],
        "content_type": [not_empty, unicode_safe],
        "size": [not_empty, int_validator],
        "hash": [default(""), unicode_safe],
        "sample": [ignore_missing, files_into_upload],
    }


@validator_args
def multipart_refresh(not_empty: Validator, unicode_safe: Validator) -> Schema:
    return {"id": [not_empty, unicode_safe]}


@validator_args
def multipart_update(
    not_missing: Validator,
    int_validator: Validator,
    not_empty: Validator,
    unicode_safe: Validator,
    files_into_upload: Validator,
) -> Schema:
    return {
        "id": [not_empty, unicode_safe],
        "part": [not_missing, int_validator],
        "upload": [not_missing, files_into_upload],
    }


@validator_args
def multipart_complete(not_empty: Validator, unicode_safe: Validator, boolean_validator: Validator) -> Schema:
    return {
        "id": [not_empty, unicode_safe],
        "keep_storage_data": [boolean_validator],
        "keep_plugin_data": [boolean_validator],
    }



import pytest
from faker import Faker
from responses import RequestsMock

from ckanext.files import shared


@pytest.mark.usefixtures("with_plugins")
class TestStorage:
    def test_url_is_kept_in_storage_data(self, faker: Faker):
        """Test that URL is kept in storage data.

        Given a link storage, when uploading a file with a URL,
        then the URL is kept in the storage data.
        """
        storage = shared.make_storage("test", {"type": "files:link"})

        location = shared.Location(faker.file_name())
        url = faker.url()

        with RequestsMock() as rsps:
            rsps.add("HEAD", url)
            data = storage.upload(location, shared.make_upload(url.encode()))

        assert data.location == location
        assert data.storage_data["url"] == url

    def test_permanent_link_returns_url(self, faker: Faker):
        """Test that permanent link returns the URL.

        Given a link storage, when getting the permanent link of a file,
        then the URL is returned.
        """
        storage = shared.make_storage("test", {"type": "files:link"})

        url = faker.url()
        data = shared.FileData(
            shared.Location(faker.file_name()),
            storage_data={"url": url},
        )

        permanent_link = storage.permanent_link(data)

        assert permanent_link == url

    def test_protocols_setting(self, faker: Faker):
        """Test that protocols setting is respected.

        Given a link storage with custom protocols, when uploading a file with a URL,
        then the URL is accepted only if it uses one of the allowed protocols.
        """
        storage = shared.make_storage("test", {"type": "files:link", "protocols": ["https"]})

        location = shared.Location("file.txt")

        url = faker.url(["http"])
        with pytest.raises(shared.exc.ContentError):  # noqa: PT012, SIM117
            storage.upload(location, shared.make_upload(url.encode()))

        url = faker.url(["https"])
        with RequestsMock() as rsps:
            rsps.add("HEAD", url)
            data = storage.upload(location, shared.make_upload(url.encode()))

        assert data.location == location
        assert data.storage_data["url"] == url

    def test_domains_setting(self, faker: Faker):
        """Test that domains setting is respected.

        Given a link storage with custom domains, when uploading a file with a URL,
        then the URL is accepted only if it uses one of the allowed domains.
        """
        allowed_domain = "example.com"
        storage = shared.make_storage("test", {"type": "files:link", "domains": [allowed_domain]})

        location = shared.Location("file.txt")

        url = f"https://notallowed.com/{faker.file_name()}"
        with pytest.raises(shared.exc.ContentError):  # noqa: PT012, SIM117
            storage.upload(location, shared.make_upload(url.encode()))

        url = f"https://{allowed_domain}/{faker.file_name()}"
        with RequestsMock() as rsps:
            rsps.add("HEAD", url)
            data = storage.upload(location, shared.make_upload(url.encode()))

        assert data.location == location
        assert data.storage_data["url"] == url



from __future__ import annotations

import base64
import dataclasses
import re
from typing import Any

from file_keeper.default.adapters import gcs
from typing_extensions import override

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.config.declaration import Declaration, Key

from ckanext.files import shared

RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")
HTTP_RESUME = 308


def decode(value: str) -> str:
    return base64.decodebytes(value.encode()).hex()


class Reader(shared.Reader, gcs.Reader):
    @override
    def response(self, data: shared.FileData, extras: dict[str, Any]) -> types.Response:
        return tk.redirect_to(self.temporary_link(data, 60, extras))


@dataclasses.dataclass()
class Settings(shared.Settings, gcs.Settings):
    pass


class GoogleCloudStorage(shared.Storage, gcs.GoogleCloudStorage):  # pyright: ignore[reportIncompatibleVariableOverride]
    hidden = True

    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = Settings
    ReaderFactory = type("Reader", (shared.Reader, gcs.GoogleCloudStorage.ReaderFactory), {})
    ManagerFactory = type("Manager", (shared.Manager, gcs.GoogleCloudStorage.ManagerFactory), {})
    UploaderFactory = type("Uploader", (shared.Uploader, gcs.GoogleCloudStorage.UploaderFactory), {})

    @override
    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)
        declaration.declare(key.bucket).required().set_description(
            "Name of the GCS bucket where uploaded data will be stored.",
        )
        declaration.declare(key.credentials_file).set_description(
            "Path to the credentials file used for authentication by GCS client."
            + "\nIf empty, uses value of GOOGLE_APPLICATION_CREDENTIALS envvar.",
        )
        declaration.declare(
            key.resumable_origin,
            tk.config["ckan.site_url"],
        ).set_description(
            "Value of the Origin header set for resumable upload."
            + "\nIn most cases, keep it the same as CKAN base URL.",
        )



"""Link storage implementation for CKAN file storage extension.

This storage backend allows CKAN to handle files that are represented by URLs.
It provides capabilities to analyze link metadata, create file entries from links,
and generate permanent links to the files.
"""

from __future__ import annotations

import dataclasses
import logging
from typing import Any
from urllib.parse import urlparse, urlunparse

import requests
from typing_extensions import override

from ckanext.files import shared

log = logging.getLogger(__name__)


class Reader(shared.Reader):
    capabilities = shared.Capability.LINK_PERMANENT
    storage: LinkStorage

    @override
    def permanent_link(self, data: shared.FileData, extras: dict[str, Any]) -> str:
        return data.storage_data.get("url", data.location)


class Uploader(shared.Uploader):
    capabilities = shared.Capability.CREATE
    storage: LinkStorage

    @override
    def upload(
        self,
        location: shared.Location,
        upload: shared.Upload,
        extras: dict[str, Any],
    ) -> shared.FileData:
        try:
            parsed = urlparse(upload.stream.read().decode())
        except ValueError as err:
            raise shared.exc.ContentError(self, str(err)) from err

        protocols = self.storage.settings.protocols
        if protocols and parsed.scheme not in protocols:
            msg = f"{parsed.scheme} protocol is not supported"
            raise shared.exc.ContentError(self, msg)

        domains = self.storage.settings.domains
        if domains and parsed.hostname not in domains:
            msg = f"Domain {parsed.hostname} is not allowed"
            raise shared.exc.ContentError(self, msg)

        url = urlunparse(parsed)
        resp = requests.head(url, timeout=self.storage.settings.timeout)
        if not resp.ok:
            log.debug("Cannot analyze URL %s: %s", url, resp)

        content_length = resp.headers.get("content-length") or "0"
        size = int(content_length) if content_length.isnumeric() else 0

        content_type = resp.headers.get("content-type") or "application/octet-stream"
        content_type = content_type.split(";", 1)[0]

        hash = resp.headers.get("etag") or ""

        return shared.FileData(location, size=size, content_type=content_type, hash=hash, storage_data={"url": url})


@dataclasses.dataclass()
class Settings(shared.Settings):
    timeout: int = 5
    protocols: list[str] = dataclasses.field(default_factory=list)
    domains: list[str] = dataclasses.field(default_factory=list)


class LinkStorage(shared.Storage):
    hidden = True
    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ReaderFactory = Reader

    @override
    @classmethod
    def declare_config_options(
        cls,
        declaration: shared.types.Declaration,
        key: shared.types.Key,
    ):
        super().declare_config_options(declaration, key)
        declaration.declare_int(key.timeout, Settings.timeout).set_description(
            "Request timeout used when link details fetched during upload.",
        )
        declaration.declare_list(key.protocols, None).set_description(
            "List of allowed protocols for link uploads. Empty list means all protocols are allowed.",
        )
        declaration.declare_list(key.domains, None).set_description(
            "List of allowed hostnames for link uploads. Empty list means all domains are allowed.",
        )
